/*

 DESCRIPTION:
 GPIBCON is a simple utility that provides a terminal-based console for GPIB
 instruments communicating over a virtual serial port.  The display is managed
 with simple ANSI cursor control sequences.  An ASCII configuration file is used
 to define the commands available along with startup, shutdown, and read.  See
 the configuration file for details.

 By default the config file is opened in the same path as the program with the
 '.conf' extension.  An alternate config file can be specified to manage
 multiple instruments.

 Displayed data can also be logged (written) to a file specified with the
 proper option on the command line.  The file is formatted as an ascii csv
 with the first row containing field names (e.g. the display labels).  Each
 record includes the date & time it was written.

 The program is written in ANSI C and should compile without modification on most
 U/L*nux & mac OS platforms.  Provided the include file is in the same folder
 the command 'cc gpibcon.c -lpthread' will create the executable.

 GPIBCON communicates using a virtual serial port enumerated as a character
 device in /dev.

 This software is licensed under the MIT license:

 Copyright 2021 Brian Cornell.

 Permission is hereby granted, free of charge, to any person obtaining a copy of
 this software and associated documentation files (the "Software"), to deal in
 the Software without restriction, including without limitation the rights to
 use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 of the Software, and to permit persons to whom the Software is furnished to do
 so, subject to the following conditions:

 The above copyright notice and this permission notice shall be included in all
 copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

 SYNTAX:        run at command line without options.

 KNOWN BUGS:    1) Command prompt will appear to hang if a command is issued
                from the config file expecting a response when in fact the
                instrument does not send one.

 DEPENDENCIES:  gpibcon.h - program include file

 VERSION:  5.1

 MODIFICATION HISTORY:

 01 Mar 21, BWC, minor revision 5.1:
    + Added ability to escape semicolon within fields so they can be used as
      part of instrument command syntax. All modifications contained in
      function getCfgVar. To escape a semicolon for use as part of command
      string place a backslash in front of it. Example: the below command for
      the Keysight 3646a PS selects output one and programs the voltage for
      5.0V.
        inst:nsel 1;:volt 5.0
        - The semicolon must be escaped in the config file to avoid being
          interpreted as a field delimiter: inst:nsel 1\;:volt 5.0.
      A backslash placed before any other character will not be treated as an
      escape and will be passed literally in the field string. So the above
      command string with a backslash in the middle of the word volt
      (inst:nsel 1;:vo\lt 5.0) will be passed to the instrument as 'vo\lt'.
    + Implemented termios to disable local echo and turn off all kernel tty/
      serial port features that implement flow control or character translation.
      Hence, data is passed thru without manipulation or HW/SW flow control. An
      ommission since v1, it went undiscovered because problems didn't manifest
      with the instruments being used. It was discovered when working with a
      Keysight 3646a PS and Wireshark to sniff the USB data. The termios
      functions tcgetattr & tcsetattr are used to retrieve & modify the tty
      characteristics respectively. The Linux specific implementation of termios
      is used and may require additional work to compile with other U*nix OS.

 10 Oct 19, BWC, fifth release:
    + Change default display timing flag - added '-u' flag.
    + Run commands in display that don't display but still log by setting
      <row> to zero.
    + Clear receive buffer after initialization to prevent display data sync
      problems:  gpibInit repetatively reads stream after each 'basic' command.
    + Clear receive buffer after off commands to prevent subsequent read
      problems:  gpibInit repetatively reads stream after each OFF command.
    + Added flag to force the reset of displayed data by purging any pending
      data in the stream at start of display sequence:  '-f' flag.
    + Added flag to reset displayed data when an I/O error was encountered on
      the last display loop by purging any pending data in the stream at start
      of display sequence:  '-r' flag.
    + Added Mark Comment field when logging.  Typing text after the mark command
      will be placed in the Comment field of the record where the mark value
      changes.  It is enclosed in quotes to accommodate punctuation that might
      be interpreted as a field separator.
    + Added record numbers to log file to make it easier to identify 'missing'
      records during analysis.
    + Added option to log config (console) commands to a file.  Helpful to
      correlate configuration changes to logged data.

 18 Sep 19, BWC, maintenance release 4.1 for compatibility with Ubuntu 18.04.
 Moving from OS X / BSD found several string functions are not natively
 supported and require libbsd.  The library libbsd runtime is distributed with
 the base system but the includes for development are not.  libbsd-dev must
 be installed on the system.
    + This release should compile properly on Darwin, BSD, and most mainstream
      linux systems.
    + To compile on linux, assuming libbsd support exists:
        gcc -pthread gpibcon.c -lbsd
    + Changed NULL string assignment from system NULL to CHR_NULL to rectify
      compiler warnings.
    + Added hack to function gpibRead to eliminate NULL read responses (only
      containing NL) that caused NULL string displays.  This only exhibits on
      Ubuntu (other linux unknown?) systems; not OS X (Darwin) or BSD.
      Basically, when reading an expected string response after issuing the read
      command to the GPIB controller a NL is returned.  A subsequent read
      (without issuing another read command) returns the expected response.  Was
      not able to determine if this is a stream I/O or kernel setting.

 20 Apr 19, BWC, 4.0 release:
    + Added ability to address different GPIB instruments on a given device:
      created the 'addr:' statement type to accomodate.
    + Corrected logic in function loadCfg to allow empty config file segments.
    + Added logic to display & init functions to not attempt to issue read
      command if not defined.
    + Changed GPIB device streams from blocking to non-blocking and added device
      timeout parameter to 'read:' records so program doesn't hang when no data
      received from device when expected.
    + Consolidated GPIB device read operations into function gpibRead to support
      non-blocking operation:  data is received in fragments and must be
      concatenated for a valid response.
    + Added precision field to display records to control the decimal resolution.
    + Increased static size of dictionary from 50 to 500 objects
    + Added delay feature to 'init:' records
    + Added command response compare delay to init records
    + Added progress indicator as init commands are processed; useful when
      loading a large instrument config
    + Added 'spinning' progress indicator to console display so the user knows
      the display is frozen.
    + Added support for optional second GPIB device channel.  All display fields
      are managed in one linked list so calculated fields work across both.
    + Added 'mark' internal console command that sets / clear integer field in
      data log as a convenience for user to filter areas of interest during
      analysis.
    + Optimization of function loadCfg.
    + Optimization of I/O operataions across numerous functions with a focus on
      preventing I/O on non-existent stream (which leads to a core dump).

 29 Mar 19, BWC, third release.  Added logging of displayed data to file and
 ability for calculated fields to use constants.

 24 Mar 19, BWC, second release.  Added calculated field capability.  Refer to
 GPIBCON.CONF for details.

 13 Feb 19, BWC, first release.


 */


#if defined(__linux__)
#include <bsd/stdio.h>
#include <bsd/stdlib.h>
#include <bsd/string.h>
#include <bsd/unistd.h>
#else
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <termios.h>
#include <stdint.h>
#include <strings.h>
#include <time.h>
#include <errno.h>
#include <signal.h>
#include <limits.h>
#include <pthread.h>

#include "gpibcon.h"


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * Module data
 +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
FILE *GpibDev[SIZ_ARY_DEV] = {NULL,NULL},   //GPIB device streams
     *LogFile = NULL;                       //Log file handle
struct termios GpibTTY[SIZ_ARY_DEV];        //GPIB tty settings
gcmd *Index[SIZ_ARY_IDX];                   //Index of first / last pointers
gcmd Dict[SIZ_ARY_DCT],                     //Command dictionary
     Read_gpib[SIZ_ARY_DEV],                //Read commands for each device
     Addr_gpib[SIZ_ARY_ADR];                //Address commands for each device
const char *prgName = STR_PRGNAME,
            *tag_calc = DSP_TAG_CALC,
            *tag_delay = INT_TAG_DELAY,
            *ary_math = MATH_ARY_OPS;
char gpibDspMarkText[SIZ_BUF_CON],          //Buffer for mark comment
    *Logc_fp = NULL;                        //Pointer to command log file path
int gpibDspWait = FALSE,                    //gpibDsp process control
    gpibDspAck = FALSE,                     //gpibDsp process control
    gpibDspMark = FALSE,                    //Log file record default 'mark' value
    flagDspForce = FALSE,                   //TRUE=display stream always resets
    flagDspReset = FALSE;                   //TRUE=display stream resets on ERR
struct timespec update_tm;                  //Display update rate (gpibDsp)


/*Function main.  Controls program execution - initialization, run, shutdown.
 *  Spawns a thread to display console data (uses pthread).
 *
 * Returns:         return code; refer to 'RC_' definitions in include file.
 *
 * Parameters:      argc=argument count
 *                  argv=array of pointers to argument strings
 *
 */
int main(int argc, char *argv[])
{
    gcmd *dsp_p;
    pthread_t id;
    int ReturnCode,
        counter,                            //Argument flag counter
        argc_c,                             //Argument counter
        argc_d,                             //argv element for device
        cflags;

    //Set default display update rate for function gpibDsp
    update_tm.tv_sec = DLY_S_DUP_C;
    update_tm.tv_nsec = DLY_NS_DUP_C;

    //Ensure buffer is NULL string
    gpibDspMarkText[0] = CHR_NULL;

    while (TRUE)
    {
        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * Argument processing:  If no arguments display program info & exit.
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        ReturnCode = RC_NOARG;
        if (argc < SIZ_ARGC_MIN)
        {
            printf(FMT_PRG_DESC, VERMAJOR, VERMINOR, VERBUILD);
            printf(FMT_PRG_SYNTX, prgName);
            for (counter = 0; (counter < SIZ_A_FLG); ++counter)\
                printf(FMT_PRG_CMDL, Flags[counter].str, Flags[counter].desc);

            break;
        }

        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * Parse command line flags.  The outer loop walks thru the list of
         * arguments and the inner loop scans the array of Flags against each
         * argument.  When a flag is found its respective bit in cflags is set
         * and the value pointer in Flags is set to that parameter (beginning
         * with the flag and not the start of the value itself).
         *
         * Invalid flags are ignored.
         *
         * When the outer loop is complete argc_d represents the offset to the
         * last flag found.  This is used when time to open the GPIB device
         * streams to determine the number of remaining command line parameters.
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        for (argc_d = 0, argc_c = 1, cflags = 0; (argc_c < argc); ++argc_c)
        {
            for (counter = 0; (counter < SIZ_A_FLG); ++counter)
                if (!(strncasecmp(Flags[counter].str, argv[argc_c], SIZ_CMD_FLAG)))
                {
                    Flags[counter].value = argv[argc_c];
                    if (argc_c > argc_d) argc_d = argc_c;
                    break;
                }
            if (counter < SIZ_A_FLG) cflags |= (FLG_BIT_SET << counter);
        }

        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * Check for display reset flags & set.
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
         if (cflags & FLG_FORCE_DSP) flagDspForce = TRUE;
         if (cflags & FLG_RESET_DSP) flagDspReset = TRUE;

        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * Load config file.  If the 'different' flag is set skip past the
         * flag portion of the value; otherwise set value to the program name.
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        if (cflags & FLG_DIFF_CONF) Flags[FLG_A_DIFF_CONF].value += 2;
        else Flags[FLG_A_DIFF_CONF].value = argv[POS_ARG_PRG];
        if ((ReturnCode = loadCfg(Flags[FLG_A_DIFF_CONF].value, \
                                  (cflags & FLG_DIFF_CONF), Index, Dict, \
                                  Read_gpib, Addr_gpib))) break;

        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * If dump flag set, dump config & exit
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        if (cflags & FLG_DUMP_CONF)
        {
            dumpCfg(Index, Dict, Read_gpib, Addr_gpib);
            ReturnCode = RC_SUCCESS;
            break;
        }

        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * If update flag set, change default display update rate
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
         if (cflags & FLG_DSP_UPDATE)
         {
             Flags[FLG_A_DSP_UPDATE].value += 2;
             update_tm.tv_sec = (strtol(Flags[FLG_A_DSP_UPDATE].value, NULL, \
                 BASE_TEN) / 1000);
             update_tm.tv_nsec = ((strtol(Flags[FLG_A_DSP_UPDATE].value, NULL, \
                 BASE_TEN) % 1000) * 1000000);
         }

        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * If log data flag set open file and write header (list of display
         * fields).
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        if (cflags & FLG_LOG_FILE)
        {
            ReturnCode = RC_LOGFAIL;
            Flags[FLG_A_LOG_FILE].value += strlen(Flags[FLG_A_LOG_FILE].str);
            if ((LogFile = fopen(Flags[FLG_A_LOG_FILE].value, \
                                 FMODE_LOG_FIL)) == NULL) break;

            if (!(fprintf(LogFile, FMT_LOG_HDR))) break;
            for (dsp_p = Index[DICT_DISP_F]; (dsp_p); dsp_p = dsp_p->next)
                fprintf(LogFile, FMT_LOG_FLD, dsp_p->cfg);
        }

        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * Initilize command log if log cfg (console) commands flag set.
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        if (cflags & FLG_CFG_WRITE)
        {
            ReturnCode = RC_CMDFAIL;
            Logc_fp = Flags[FLG_A_CFG_WRITE].value + \
                strlen(Flags[FLG_A_CFG_WRITE].str);
            if (!(gpibLogc(Logc_fp, NULL, NULL, NULL))) break;
        }

        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * Open GPIB device:  argc_d is expected to be set to element of last
         * flag (or first parameter after program name if none).  Return code is
         * set to BADARG and argc_d incremented & checked against parameter
         * count.  Exit program if no more command line parameters.
         *
         * Otherwise, reset return code to DEVOFAIL to indicate problem opening
         * stream and attempt to open first device.  Second device is optional -
         * increment argc_d again and attempt open if valid.
         *
         * The TTY attributes are retrieved, canonical & echo modes are
         * disabled, and written back to the port. The devices are then set to
         * non-blocking so the program doesn't hang if a response isn't received
         * when expected.  Hence, functions communicating with the devices
         * accommodate this condition.
         *
         * As of v5.1 fcntl is still used to set non-blocking but this Should
         * be changed in future versions to be handled by tcsetattr for
         * consistency and more granular control.
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        ReturnCode = RC_BADARG;
        if (++argc_d == argc) break;
        ReturnCode = RC_DEVOFAIL;
        if ((GpibDev[ARY_FLD_DEV1] = fopen(argv[argc_d], FMODE_GPIB)) == NULL)
            break;
        tcgetattr(fileno(GpibDev[ARY_FLD_DEV1]), &GpibTTY[ARY_FLD_DEV1]);
        GpibTTY[ARY_FLD_DEV1].c_lflag &= ~ICANON;       //Disable canonical mode
        GpibTTY[ARY_FLD_DEV1].c_lflag &= ~ECHO;         //Disable echo
        GpibTTY[ARY_FLD_DEV1].c_lflag &= ~ECHOE;        //Disable erasure
        GpibTTY[ARY_FLD_DEV1].c_lflag &= ~ECHONL;       //Disable new-line echo
        GpibTTY[ARY_FLD_DEV1].c_cflag &= ~CRTSCTS;      //Disable RTS/CTS
        GpibTTY[ARY_FLD_DEV1].c_cflag |= CREAD | CLOCAL; //Turn on READ & ignore ctrl lines
        GpibTTY[ARY_FLD_DEV1].c_lflag &= ~ISIG;         //Disable interpretation of INTR, QUIT and SUSP
        GpibTTY[ARY_FLD_DEV1].c_iflag &= ~(IXON | IXOFF | IXANY);   //Turn off s/w flow ctrl
        GpibTTY[ARY_FLD_DEV1].c_iflag &= ~(IGNBRK|BRKINT|PARMRK|ISTRIP|INLCR|IGNCR|ICRNL);  //Disable any special handling of received bytes
        GpibTTY[ARY_FLD_DEV1].c_oflag &= ~OPOST;        //Prevent special interpretation of output bytes (e.g. newline chars)
        GpibTTY[ARY_FLD_DEV1].c_oflag &= ~ONLCR;        //Prevent conversion of newline to carriage return/line feed
        tcsetattr(fileno(GpibDev[ARY_FLD_DEV1]), TCSANOW, \
            &GpibTTY[ARY_FLD_DEV1]);
        fcntl(fileno(GpibDev[ARY_FLD_DEV1]), F_SETFL, O_NONBLOCK);

        if (++argc_d < argc)
        {
            if ((GpibDev[ARY_FLD_DEV2] = fopen(argv[argc_d], \
                                               FMODE_GPIB)) == NULL) break;
            tcgetattr(fileno(GpibDev[ARY_FLD_DEV2]), &GpibTTY[ARY_FLD_DEV2]);
            GpibTTY[ARY_FLD_DEV2].c_lflag &= ~ICANON;   //Disable canonical mode
            GpibTTY[ARY_FLD_DEV2].c_lflag &= ~ECHO;     //Disable echo
            GpibTTY[ARY_FLD_DEV2].c_lflag &= ~ECHOE;    //Disable erasure
            GpibTTY[ARY_FLD_DEV2].c_lflag &= ~ECHONL;   //Disable new-line echo
            GpibTTY[ARY_FLD_DEV2].c_cflag &= ~CRTSCTS;      //Disable RTS/CTS
            GpibTTY[ARY_FLD_DEV2].c_cflag |= CREAD | CLOCAL; //Turn on READ & ignore ctrl lines
            GpibTTY[ARY_FLD_DEV2].c_lflag &= ~ISIG;         //Disable interpretation of INTR, QUIT and SUSP
            GpibTTY[ARY_FLD_DEV2].c_iflag &= ~(IXON | IXOFF | IXANY);   //Turn off s/w flow ctrl
            GpibTTY[ARY_FLD_DEV2].c_iflag &= ~(IGNBRK|BRKINT|PARMRK|ISTRIP|INLCR|IGNCR|ICRNL);  //Disable any special handling of received bytes
            GpibTTY[ARY_FLD_DEV2].c_oflag &= ~OPOST;        //Prevent special interpretation of output bytes (e.g. newline chars)
            GpibTTY[ARY_FLD_DEV2].c_oflag &= ~ONLCR;        //Prevent conversion of newline to carriage return/line feed
            tcsetattr(fileno(GpibDev[ARY_FLD_DEV2]), TCSANOW, \
                &GpibTTY[ARY_FLD_DEV2]);
            fcntl(fileno(GpibDev[ARY_FLD_DEV2]), F_SETFL, O_NONBLOCK);
        }

        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * Once opened assign the stream pointers to their respective dictionary
         * records.  To avoid a pointer overflow or misassigned stream devn
         * must be less than the size of the GPIB device stream array
         * (SIZ_ARY_DEV).  Otherwise devf is left unassigned and it's expected
         * that function loadCfg initialized the dictionary to NULLs so they
         * will contain an invalid stream handle that subsequent functions can
         * test or ignore without damage.
         *
         * Address & read records are static assignments to separate dictionary
         * array.
         *
         * Note:  config file syntax labels devices starting with offset 1.
         * The config loader must addjust to offset 0.
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        Read_gpib[ARY_FLD_DEV1].devf = GpibDev[ARY_FLD_DEV1];
        Read_gpib[ARY_FLD_DEV2].devf = GpibDev[ARY_FLD_DEV2];
        //Init records
        for (dsp_p = Index[DICT_INIT_F]; (dsp_p); dsp_p = dsp_p->next)
            if (dsp_p->devn <  SIZ_ARY_DEV) dsp_p->devf = GpibDev[dsp_p->devn];
        //Config records
        for (dsp_p = Index[DICT_CONF_F]; (dsp_p); dsp_p = dsp_p->next)
            if (dsp_p->devn <  SIZ_ARY_DEV) dsp_p->devf = GpibDev[dsp_p->devn];
        //Display records
        for (dsp_p = Index[DICT_DISP_F]; (dsp_p); dsp_p = dsp_p->next)
            if (dsp_p->devn <  SIZ_ARY_DEV) dsp_p->devf = GpibDev[dsp_p->devn];
        //Off records
        for (dsp_p = Index[DICT_OFF_F]; (dsp_p); dsp_p = dsp_p->next)
            if (dsp_p->devn <  SIZ_ARY_DEV) dsp_p->devf = GpibDev[dsp_p->devn];

        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * Install signal handlers
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        ReturnCode = RC_SIGERROR;
        if (signal(SIGINT, sigHandler) == SIG_ERR) break;
        if (signal(SIGTERM, sigHandler) == SIG_ERR) break;
        if (signal(SIGHUP, sigHandler) == SIG_ERR) break;
        if (signal(SIGQUIT, sigHandler) == SIG_ERR) break;

        //Initialize GPIB interface
        ReturnCode = RC_INITFAIL;
        if (gpibInit(Index[DICT_INIT_F], Read_gpib)) break;

        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * Setup display and start display thread
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        scrSetup();
        ReturnCode = RC_THREAD;
        if (pthread_create(&id, NULL, (void *(*)(void *)) &gpibDsp, NULL)) break;

        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * Enter console mmode; exit upon return.
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        ReturnCode = cmdProc();
        break;
    }

    //Shutdown
    if (GpibDev[ARY_FLD_DEV1])
    {
        gpibOff(Index[DICT_OFF_F]);
        fclose(GpibDev[ARY_FLD_DEV1]);
    }
    if (GpibDev[ARY_FLD_DEV2]) fclose(GpibDev[ARY_FLD_DEV2]);
    if (LogFile) fclose(LogFile);

    //Exit messages
    if (ReturnCode > RC_NOARG) fprintf(stderr, FMT_ERR_STDERR, prgName,\
                                       retCode[ReturnCode], errno);
    printf(FMT_PRG_EXIT);

    return(ReturnCode);
} //End function main


/*Function cmdProc.  Console command processor.  Relies on access to module
 *  data.
 *
 * Returns:         non-zero on error
 *
 * Parameters:
 *
 */
int cmdProc(void)
{
    gcmd *req;                                  //User command request
    struct timespec base_to;                    //Used for nanosleep
    char console[SIZ_BUF_CON],
        req_c[SIZ_BUF_CON],                     //Command portion of request
        req_v[SIZ_BUF_CON],                     //Value portion of request
        *fgets_p,                               //Function fgets return value
        Logc_noerr[SIZ_ARY_NULM] = {0x00, 0x00},    //Null msg for no Logc error
        *Logc_err;                              //Pointer to gpibLogc error msg
    int int_c,                                  //Internal command array offset
        rc = RC_SUCCESS,                        //Return code
        loop_c;                                 //Loop counter

    base_to.tv_sec = DLY_S_RD_C;                //Delay between GPIB commands
    base_to.tv_nsec = DLY_NS_RD_C;

    while (TRUE)
    {
        //Setup command line
        memset((void *) console, CHR_NULL, SIZ_BUF_CON);
        ansiCursorPos(DSP_POS_CMD_R, DSP_POS_CMD_C);
        printf(FMT_CMD_PR);
        ansiEraseLine();
        fflush(stdout);
        fgets(console, SIZ_BUF_CON, stdin);

        //Setup display for response
        ansiCursorPos(DSP_POS_RSP_R, DSP_POS_RSP_C);
        printf(FMT_RSP_PR);
        ansiEraseLine();
        fflush(stdout);

        //Parse command line
        strnl((unsigned char *) console, SIZ_BUF_CON);      //Strip CR/LF
        strlcpy(req_c, console, SIZ_BUF_CON);               //Copy for command
        strlrtrim((unsigned char *) req_c, SIZ_BUF_CON);    //Strip trailing text
        strlcpy(req_v, console, SIZ_BUF_CON);               //Copy for value
        strlnshift((unsigned char *) req_v, SIZ_BUF_CON);   //Strip leading text

        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * Process matching commands in dictionary
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        for (req = Index[DICT_CONF_F]; (req); req = req->next)
            if (!(strcasecmp(req_c, req->cfg))) break;

        if (req)
        {
            //Display error if value required with command but none
            if ((req->usrv) && (!(strlen(req_v)))) printf(FMT_RSP_NOV);
            //Valid command - process
            else
            {
                //Pause display thread so no GPIB device contention
                gpibDspWait = TRUE;
                while (!(gpibDspAck));          //Wait until released

                //Set device address & write command if valid device
                setDevAddr(req);
                if (req->devf) fprintf(req->devf, FMT_EXC_CMD, req->cmd, req_v);
                if (req->crsp)                  //Only get response if expecting
                {
                    nanosleep(&base_to, NULL);
                    //Skip read command if not defined
                    if (strlen(Read_gpib[req->devn].cmd)) \
                        fprintf(req->devf, FMT_DSP_WR, Read_gpib[req->devn].cmd);

                    //Read command response, do not remove leading text
                    gpibRead(req, 0);
                }
                else *req->rsp = CHR_NULL;

                //Print result & resume display thread
                Logc_err = Logc_noerr;
                if (Logc_fp)
                    if (!(gpibLogc(Logc_fp, req->cmd, req_v, req->rsp)))
                        Logc_err = (char *) retCode[RC_CMDFAIL];
                printf(FMT_EXC_RSP, req->rsp, Logc_err);
                gpibDspWait = FALSE;
            }
        }
        else
        {
            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             * Process internal commands
             +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
            for (int_c = 0; (int_c < SIZ_A_ICMD); ++int_c)
                if (!(strcasecmp(req_c, Icmd[int_c].str))) break;

            if (int_c == CMD_A_QUIT)
            {
                ansiCursorPos(DSP_POS_EXT_R, DSP_POS_EXT_C);
                break;
            }
            if (int_c == CMD_A_MARK)
            {
                gpibDspMark ^= TRUE;
                strlcpy(gpibDspMarkText, req_v, SIZ_BUF_CON);
            }
            else if (int_c == CMD_A_CLS)
            {
                gpibDspWait = TRUE;
                while (!(gpibDspAck));                  //Wait until released
                scrSetup();
                gpibDspWait = FALSE;                    //Resume
            }
            else if (int_c == CMD_A_HELP)
            {
                gpibDspWait = TRUE;
                while (!(gpibDspAck));                  //Wait until released
                ansiClrScr();
                printf(FMT_HLP_DCM);

                //Display dictionary help
                for (req = Index[DICT_CONF_F]; (req); req = req->next)
                    printf(FMT_HLP_TXT, req->cfg, req->hlp);

                //Display internal help
                printf(FMT_HLP_ICM);
                for (int_c = 0; (int_c < SIZ_A_ICMD); ++int_c)
                    printf(FMT_HLP_TXT, Icmd[int_c].str, Icmd[int_c].desc);
                printf(FMT_HLP_END);                    //All done
                fgets(console, SIZ_BUF_CON, stdin);     //Wait for user
                scrSetup();                             //Redraw screen
                gpibDspWait = FALSE;                    //Resume
            }
            else printf(FMT_RSP_INV);
        }
    }

    return(rc);
} //End function cmdProc


/*Function loadCfg.  Parses config file, loads dictrionary, initializes
 * pointers.  Erroneous config statements will only result in an incomplete
 * config record that will cause unintended operation but not execution failure.
 * This function does not require all dictionary segments to be loaded since
 * it's reasonable that some configurations may not require, say, commands or
 * exit statements.
 *
 * Returns:         RC_SUCCESS if loaded, otherwise RC_BADCFG (only caused by
 *                  failure to open file)
 *
 * Parameters:      cfg=config file to load
 *                  cfg_d=FALSE to load default config; otherwise load cfg
 *                      without modification
 *                  idx=array of pointers to manage dictionary linked list
 *                  dct=config dictionary array
 *                  dct_r=read command array
 *                  dct_a=address command array
 *
 */
int loadCfg(char *cfg, int cfg_d, gcmd *idx[], gcmd *dct, gcmd *dct_r, \
            gcmd *dct_a)
{
    FILE *GpibCfg;                          //Config file handle
    gcmd *currec;                           //Current dictionary record to initialize
    int NumObj,                             //Number of objects loaded
        NumObj_p,                           //Number of ojects with pointers setup
        line,                               //Config line number
        rc,                                 //Return code
        sz_init,
        sz_read,
        sz_cfg,
        sz_dsp,
        sz_off,
        sz_addr,
        first_a,                            //Dict. index to use for first ptr
        last_a;                             //Dict. index to use for last ptr
    char CfgFile[SIZ_BUF_FILE],             //Buffer for config file path
        *rec_init,
        *rec_read,
        *rec_cfg,
        *rec_dsp,
        *rec_off,
        *rec_addr;
    const char *fld_math;
    unsigned int devn_r;

    for (GpibCfg = NULL, rc = RC_BADCFG; (TRUE);)
    {
        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * Open config file & initialize memory objects
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        if (strlcpy(CfgFile, cfg, SIZ_BUF_FILE) > SIZ_BUF_FILE) break;
        if (!(cfg_d))
            if (strlcat(CfgFile, STR_EXT_CFG, SIZ_BUF_FILE) > SIZ_BUF_FILE) break;
        if (!(GpibCfg = fopen(CfgFile, FMODE_CONF))) break;

        memset((void *) dct, CHR_NULL, (SIZ_ARY_DCT * sizeof(gcmd)));
        memset((void *) dct_r, CHR_NULL, (SIZ_ARY_DEV * sizeof(gcmd)));
        memset((void *) idx, CHR_NULL, (SIZ_ARY_IDX * sizeof(gcmd *)));
        rec_init = REC_NAM_INIT;
        rec_read = REC_NAM_READ;
        rec_cfg = REC_NAM_CFG;
        rec_dsp = REC_NAM_DSP;
        rec_off = REC_NAM_OFF;
        rec_addr = REC_NAM_ADR;
        sz_init = strlen(rec_init);
        sz_read = strlen(rec_read);
        sz_cfg = strlen(rec_cfg);
        sz_dsp = strlen(rec_dsp);
        sz_off = strlen(rec_off);
        sz_addr = strlen(rec_addr);

        for (line = 0, NumObj = 0, NumObj_p = 0, rc = RC_SUCCESS; \
             ((fgets(CfgFile, SIZ_BUF_FILE, GpibCfg)) && (NumObj < SIZ_ARY_DCT)); \
             ++line)
        {
            //Get rid of trailing CR/LF
            strnl((unsigned char *) CfgFile, SIZ_BUF_FILE);

            //Set currec to next object to initialize
            currec = (dct + NumObj);

            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             * Initialization record:  three variants
             * 1) Basic init record with a command
             * 2) Delay indicated with $delay$ in command field
             * 3) Pause indicated with a string in command response field
             +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
            if (!(strncasecmp(CfgFile, rec_init, sz_init)))
            {
                currec->type = REC_TYP_INIT;
                strlcpy(currec->cmd, \
                        getCfgVar(CfgFile, POS_INT_CMDC), SIZ_BUF_CFG);

                //If init record is delay store value
                if (!(strncasecmp(currec->cmd, tag_delay, \
                                  strlen(tag_delay))))
                {
                    currec->usrv = \
                    strtol(getCfgVar(CfgFile, POS_INT_DELAY), NULL, BASE_TEN);
                    //Make sure not greater than max allowed
                    if (currec->usrv > DLY_MS_MAX_INIT) currec->usrv = 0;
                }
                //Copy anything found in command response field (crsp)
                else strlcpy(currec->cfg, \
                        getCfgVar(CfgFile, POS_INT_CRSP), SIZ_BUF_CFG);

                //Set addresses to update index ptr array & increment object count
                first_a = DICT_INIT_F;
                last_a = DICT_INIT_L;
                ++NumObj;
            }
            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             * Read records are not part of the main dictionary because only
             * one read record is allowed per device.  Each time a read record
             * is encountered it's device number is bounds checked and because
             * this is used for array references throughout the program and
             * out of bound value will terminate the load.  Config file syntax
             * stiplulates offset 1 and this is adjusted to offset 0 for array
             * use after bounds checking.
             *
             * Any number of read statements can be included in the config but
             * only the last one read for each channel is used.
             *
             * Read records do not increment NumObj and hence bypass the
             * common record pointer initialization block at the end of this
             * loop.
             +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
            else if (!(strncasecmp(CfgFile, rec_read, sz_read)))
            {
                devn_r = strtol(getCfgVar(CfgFile, POS_ALL_DEV), NULL, BASE_TEN);
                if ((devn_r > SIZ_ARY_DEV) || (!(devn_r)))
                {
                    rc = RC_BADCFG;
                    break;
                }
                --devn_r;                               //Adjust to offset 0

                //Set record type
                dct_r[devn_r].type = REC_TYP_READ;
                //Instrument address
                dct_r[devn_r].addr = strtol(getCfgVar(CfgFile, POS_ALL_ADR), \
                                            NULL, BASE_TEN);
                //Command to address instrument
                strlcpy(dct_r[devn_r].cmd,
                        getCfgVar(CfgFile, POS_REA_CMDC), SIZ_BUF_CMD);

                //Get device read timeout value
                dct_r[devn_r].usrv =
                strtol(getCfgVar(CfgFile, POS_REA_TOUT), NULL, BASE_TEN);
                //Make sure not greater than max allowed
                if (dct_r[devn_r].usrv > DLY_MS_MAX_INIT) dct_r[devn_r].usrv = 0;
            }
            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             * Address records are not part of the main dictionary because only
             * one addr record is allowed per device.  The device number is
             * bounds checked because this is used for array references
             * throughout the program and an out of bound value will terminate
             * the load.  Config file syntax stiplulates offset 1 and this is
             * adjusted to offset 0 for array use after bounds checking.
             *
             * Any number of address statements can be included in the config but
             * only the last one read for each device is used.
             *
             * Address records do not increment NumObj and hence bypass the
             * common record pointer initialization block at the end of this
             * loop.
             +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
            else if (!(strncasecmp(CfgFile, rec_addr, sz_addr)))
            {
                devn_r = strtol(getCfgVar(CfgFile, POS_ALL_DEV), NULL, BASE_TEN);
                if ((devn_r > SIZ_ARY_DEV) || (!(devn_r)))
                {
                    rc = RC_BADCFG;
                    break;
                }
                --devn_r;                               //Adjust to offset 0

                //Set record type
                dct_a[devn_r].type = REC_TYP_ADR;
                //Instrument address
                dct_a[devn_r].addr = strtol(getCfgVar(CfgFile, POS_ALL_ADR), \
                                            NULL, BASE_TEN);
                //Command to address instrument
                strlcpy(dct_a[devn_r].cmd,
                        getCfgVar(CfgFile, POS_ADR_CMDC), SIZ_BUF_CMD);
            }
            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             * Config (command) record
             +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
            else if (!(strncasecmp(CfgFile, rec_cfg, sz_cfg)))
            {
                currec->type = REC_TYP_CFG;
                currec->usrv = \
                strtol(getCfgVar(CfgFile, POS_CFG_PARM), NULL, BASE_TEN);
                currec->crsp = \
                strtol(getCfgVar(CfgFile, POS_CFG_CRSP), NULL, BASE_TEN);
                strlcpy(currec->cfg, \
                        getCfgVar(CfgFile, POS_CFG_USRC), SIZ_BUF_CFG);
                strlcpy(currec->cmd, \
                        getCfgVar(CfgFile, POS_CFG_CMDC), SIZ_BUF_CMD);
                strlcpy(currec->hlp, \
                        getCfgVar(CfgFile, POS_CFG_HELP), SIZ_BUF_HLP);

                //Set addresses to update index ptr array & increment object count
                first_a = DICT_CONF_F;
                last_a = DICT_CONF_L;
                ++NumObj;
            }
            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             * Display record
             +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
            else if (!(strncasecmp(CfgFile, rec_dsp, sz_dsp)))
            {
                currec->type = REC_TYP_DSP;
                currec->usrv = \
                strtol(getCfgVar(CfgFile, POS_DSP_ROW), NULL, BASE_TEN);
                currec->crsp = \
                strtol(getCfgVar(CfgFile, POS_DSP_RLTX), NULL, BASE_TEN);
                strlcpy(currec->cfg, \
                        getCfgVar(CfgFile, POS_DSP_LABL), SIZ_BUF_CFG);

                strlcpy(currec->cmd, \
                        getCfgVar(CfgFile, POS_DSP_CMDC), SIZ_BUF_CMD);

                /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
                 * If calculated field is indicated by cmd then populate math
                 * fields; otherwise math is left set to zero which which
                 * indicates a non-calculated display field.
                 *
                 * This means that if the calculated field identifier is used in
                 * the cmd field but an invalid operand is placed in the math
                 * field that the instrument will likely throw an error and
                 * the display will not render correctly.
                 *
                 * Constants are indicated by the display number field beginning
                 * with an upper/lower case 'C'.  The fld element is left set to
                 * zero and the respective fvar element is loaded with the
                 * [expected] number that follows.
                 +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
                if (!(strncasecmp(currec->cmd, tag_calc, strlen(tag_calc))))
                {
                    fld_math = getCfgVar(CfgFile, POS_DSP_MATH);
                    if (*fld_math == *(ary_math + MATH_OP_ADD))
                        currec->math = MATH_OP_ADD;
                    else if (*fld_math == *(ary_math + MATH_OP_SUB))
                        currec->math = MATH_OP_SUB;
                    else if (*fld_math == *(ary_math + MATH_OP_MUL))
                        currec->math = MATH_OP_MUL;
                    else if (*fld_math == *(ary_math + MATH_OP_DIV))
                        currec->math = MATH_OP_DIV;
                    else currec->math = MATH_OP_BAD;

                    //Check field 1 for constant
                    fld_math = getCfgVar(CfgFile, POS_DSP_FLD1);
                    if ((*fld_math == CHR_C_LC) || (*fld_math == CHR_C_UC))
                        currec->fval[ARY_FLD_FLD1] = \
                        strtof((fld_math + 1), NULL);
                    else currec->fld[ARY_FLD_FLD1] = \
                        strtol(fld_math, NULL, BASE_TEN);

                    //Check field 2 for constant
                    fld_math = getCfgVar(CfgFile, POS_DSP_FLD2);
                    if ((*fld_math == CHR_C_LC) || (*fld_math == CHR_C_UC))
                        currec->fval[ARY_FLD_FLD2] = \
                        strtof((fld_math + 1), NULL);
                    else currec->fld[ARY_FLD_FLD2] = \
                        strtol(fld_math, NULL, BASE_TEN);

                    //Get precision
                    currec->prec = \
                    strtol(getCfgVar(CfgFile, POS_DSP_PREC), NULL, BASE_TEN);
                }

                strlcpy(currec->hlp, \
                        getCfgVar(CfgFile, POS_DSP_HELP), SIZ_BUF_HLP);

                //Set addresses to update index ptr array & increment object count
                first_a = DICT_DISP_F;
                last_a = DICT_DISP_L;
                ++NumObj;
            }
            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             * Off (shutdown) record
             +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
            else if (!(strncasecmp(CfgFile, rec_off, sz_off)))
            {
                currec->type = REC_TYP_OFF;
                strlcpy(currec->cmd, \
                        getCfgVar(CfgFile, POS_OFF_CMDC), SIZ_BUF_CMD);

                //Set addresses to update index ptr array & increment object count
                first_a = DICT_OFF_F;
                last_a = DICT_OFF_L;
                ++NumObj;
            }

            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             * Initialize record pointers.  Requires each dictionary code block
             * to set first_a & last_a to their respective dictionary index
             * pointers.  NumObj must also be incremented to indicate a valid
             * record was processed and pointers need to be set.  Pointer
             * currec must point to the record just processed.
             *
             * The devn field & position are common to all record types so it
             * is loaded, bounds checked, and offset adjusted here.  Bounds
             * checking is necessary to avoid an bad array reference.  If found
             * the config load stops and returns with error code.  The config
             * file uses offset 1 to reference the device number but array
             * references are offset 0.
             *
             * The addr field is also common to all record types so it is
             * loaded here.  No bounds checking is performed.
             *
             * This is an optimization to eliminate repetative code from each
             * block.
             +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
            if (NumObj_p != NumObj)
            {
                currec->devn = \
                strtol(getCfgVar(CfgFile, POS_ALL_DEV), NULL, BASE_TEN);
                if ((currec->devn > SIZ_ARY_DEV) || (!(currec->devn)))
                {
                    rc = RC_BADCFG;
                    break;
                }
                --currec->devn;

                //Instrument address
                currec->addr = strtol(getCfgVar(CfgFile, POS_ALL_ADR), \
                                      NULL, BASE_TEN);

                NumObj_p = NumObj;
                if (!(idx[first_a]))
                {
                    idx[first_a] = currec;
                    idx[last_a] = currec;
                }
                else
                {
                    currec->prev = idx[last_a];
                    idx[last_a]->next = currec;
                    idx[last_a] = currec;
                }
            }
        }

        if (GpibCfg) fclose(GpibCfg);
        break;
    }

    return(rc);
} //End function loadCfg


/*Function dumpCfg.  Prints config file in logical order by section to stdout.
 *
 * Returns:         n/a
 *
 * Parameters:      idx=array of linked list pointers
 *                  dct=config dictionary array
 *                  dct_r=read command array
 *                  dct_a=address command array
 *
 */
void dumpCfg(gcmd *idx[], gcmd *dct, gcmd *dct_r, gcmd *dct_a)
{
    gcmd *dsp;
    char *hdr = FMT_DMP_HDR;
    int nobj = 0,
        ndsp;

    //Initialization record
    printf(hdr, REC_NAM_INIT);
    for (dsp = idx[DICT_INIT_F]; (dsp); dsp = dsp->next, ++nobj)
        printf(FMT_DMP_INT, (dsp->devn + 1), dsp->addr, dsp->cmd, dsp->cfg, \
               dsp->usrv);

    //Read record
    printf(hdr, REC_NAM_READ);
    printf(FMT_DMP_RD, (ARY_FLD_DEV1 + 1), dct_r[ARY_FLD_DEV1].addr, \
           dct_r[ARY_FLD_DEV1].cmd, dct_r[ARY_FLD_DEV1].usrv);
    printf(FMT_DMP_RD, (ARY_FLD_DEV2 + 1), dct_r[ARY_FLD_DEV2].addr, \
           dct_r[ARY_FLD_DEV2].cmd, dct_r[ARY_FLD_DEV2].usrv);

    //Address record
    printf(hdr, REC_NAM_ADR);
    printf(FMT_DMP_ADR, (ARY_FLD_DEV1 + 1), dct_a[ARY_FLD_DEV1].addr, \
           dct_a[ARY_FLD_DEV1].cmd);
    printf(FMT_DMP_ADR, (ARY_FLD_DEV2 + 1), dct_a[ARY_FLD_DEV2].addr, \
           dct_a[ARY_FLD_DEV2].cmd);

    //Config (command) record
    printf(hdr, REC_NAM_CFG);
    for (dsp = idx[DICT_CONF_F]; (dsp); dsp = dsp->next, ++nobj)
        printf(FMT_DMP_CFG, (dsp->devn + 1), dsp->addr, dsp->cfg, dsp->cmd, \
               dsp->usrv, dsp->crsp, dsp->hlp);

    //Display record
    printf(hdr, REC_NAM_DSP);
    for (dsp = idx[DICT_DISP_F], ndsp = 1; (dsp); dsp = dsp->next, ++nobj, \
         ++ndsp)
    {
        printf(FMT_DMP_DSP, ndsp, (dsp->devn + 1), dsp->addr, dsp->cfg, \
               dsp->cmd, dsp->usrv, dsp->crsp, dsp->hlp);

        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * For simplicity the math fields are printed without discrimination
         * for constants or display.  Each field pair - display # & constant -
         * are printed with an OR (vertical bar) separator and it's left to
         * the user to understand how the field will be handled:  if display
         * field is 0 then the constant value will be used.
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        if (dsp->math)
            printf(FMT_DMP_MTH, *(ary_math + dsp->math), \
                   dsp->fld[ARY_FLD_FLD1], dsp->fval[ARY_FLD_FLD1], \
                   dsp->fld[ARY_FLD_FLD2], dsp->fval[ARY_FLD_FLD2], \
                   dsp->prec);
    }

    //Off (shutdown) record
    printf(hdr, REC_NAM_OFF);
    for (dsp = idx[DICT_OFF_F]; (dsp); dsp = dsp->next, ++nobj)
        printf(FMT_DMP_OFF, (dsp->devn + 1), dsp->addr, dsp->cmd);

    printf(FMT_DMP_END, nobj, SIZ_ARY_DCT);
    return;
} //End function dumpCfg


/*Function gpibInit.  Initializes GPIB device.  Executes all variants:
 *  1) Basic
 *  2) Delay
 *  3) Wait for match
 *
 * To process the match on command response this function expects the device
 * stream to be set to non-blocking.  This is necessary for compatibility with
 * some instruments that may require repeated queries before responding due
 * to processing prior commands.  In these cases the compare will fail by
 * default since no response is loaded.
 *
 * This function will hang if the response command never responds or doesn't
 * respond with the configured text.
 *
 * Basic initialization commands (not waiting for match) will also read the
 * stream after sending the command to ensure that the stream is clear prior
 * to entering the display loop.  This prevents fields displaying out of
 * sequence.  The side effect is a longer period of time, particularly when no
 * response is received, for initialization to complete.
 *
 * Returns:     non-zero on error
 *
 * Parameters:  rec=pointer to first initialization record in dictionary
 *              dct_r=read command array
 *
 */
int gpibInit(gcmd *rec, gcmd *dct_r)
{
    struct timespec ts;

    //Setup timing
    ts.tv_sec = DLY_S_INT_C;
    ts.tv_nsec = DLY_NS_INT_C;

    for (; (rec); rec = rec->next)
    {
        //Set device address
        setDevAddr(rec);

        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        * User defined delay between commands.
        *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        if (rec->usrv)
        {
            fprintf(stdout, FMT_INT_DELAY, (rec->devn + 1), rec->usrv);
            ts.tv_nsec = (rec->usrv * DLY_INT_NS_CONV);
            nanosleep(&ts, NULL);
            ts.tv_nsec = DLY_NS_INT_C;
        }
        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        * Wait to match on command response.
        *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        else if (strlen(rec->cfg))
        {
            fprintf(stdout, FMT_INT_MATCH, rec->cmd, (rec->devn + 1), \
                    rec->addr, rec->cfg);
            while (strncasecmp(rec->rsp, rec->cfg, SIZ_BUF_CFG))
            {
                if (rec->devf) fprintf(rec->devf, FMT_INT_WR, rec->cmd);
                //Skip read command if not defined
                if (strlen(dct_r[rec->devn].cmd))
                    if (rec->devf) fprintf(rec->devf, FMT_DSP_WR, \
                                           dct_r[rec->devn].cmd);

                //Read response without stripping leading text
                gpibRead(rec, 0);
            }
        }
        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        * Just issue command.
        *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        else
        {
            fprintf(stdout, FMT_INT_CMD, rec->cmd, (rec->devn + 1), rec->addr);
            if (rec->devf) fprintf(rec->devf, FMT_INT_WR, rec->cmd);
            nanosleep(&ts, NULL);

            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
            * v4.2:  consume any response to clear stream and prevent display
            * field problems.
            *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
            while (gpibRead(rec, 0));
        }
    }

    return(RC_SUCCESS);
} //End function gpibInit


/*Function gpibOff.  Shutdown GPIB device.  Reads the stream after each
 * command to clear.
 *
 * Returns:         non-zero on error
 *
 * Parameters:  rec=pointer to first off record in dictionary
 *
 */
int gpibOff(gcmd *rec)
{
    struct timespec ts;

    //Non-blocking attempt to prevent contention if display thread is running
    gpibDspWait = TRUE;
    ts.tv_sec = DLY_S_OFT_C;
    ts.tv_nsec = DLY_NS_OFT_C;
    nanosleep(&ts, NULL);

    ts.tv_sec = DLY_S_OFF_C;
    ts.tv_nsec = DLY_NS_OFF_C;
    while (rec)
    {
        //Set device address
        setDevAddr(rec);

        if (rec->devf) fprintf(rec->devf, FMT_OFF_WR, rec->cmd);

        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        * v4.2:  consume any response to clear stream.
        *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        while (gpibRead(rec, 0));

        rec = rec->next;
        nanosleep(&ts, NULL);
    }

    return(RC_SUCCESS);
} //End function gpibOff


/*Function strnl.  Removes the trailing New Line (NL) or Carriage Return (CR)
 *  from a string up to maxsz - 1 (whichever is encountered first) and replaces
 *  with a NULL byte.  In the event that buf was not NULL terminated and NL or
 *  CR aren't found buf will be left unterminated; therefore the calling
 *  function should check the return code.
 *
 *
 * Returns:     length of NULL terminated buffer (minus NULL byte) or zero if CR
 *              or LF isn't found.  If a NULL is encountered first the function
 *              will leave the buffer unmodified and return the length of NULL
 *              terminated buffer (minus NULL byte).
 *
 * Parameters:  buf=pointer to buffer
 *              maxsz=size of buffer
 *
 */
int strnl(unsigned char *buf, int maxsz)
{
    int pos;

    for (pos = 0; (pos < maxsz); ++pos)
    {
        if ((*(buf + pos) == CHR_CR) || (*(buf + pos) == CHR_LF) || \
            (*(buf + pos) == CHR_NULL)) break;
    }
    if (pos < maxsz)
    {
        *(buf + pos) = CHR_NULL;
        return(pos);
    }
    return(0);
} //End function strnl


/*Function strlnshift.  Removes the leading text & white space from a string and
 *  shifts the remaining text left.  The shift continues until NULL or until
 *  maxsz - 1.  In the event that no text was found to shift, buf will be
 *  truncated to a NULL string.  Useful to isolate a value from a command.
 *
 * Returns:     length of left shifted text minus NULL terminator
 *
 * Parameters:  buf=pointer to string
 *              maxsz=size of buffer
 *
 */
int strlnshift(unsigned char *buf, int maxsz)
{
    int pos,
        lpos;

    //In this loop lpos is flag for leading text to eliminate
    for (pos = 0, lpos = 0; ((pos < maxsz) && (*(buf + pos))); ++pos)
    {
        if (*(buf + pos) != CHR_SPACE)
        {
            //Found text to shift; else found leading text to delete
            if (lpos == 2) break; else lpos = 1;
        }
        //Found whitespace after leading text
        else if (lpos == 1) lpos = 2;
    }

    //Shift text left; pos has starting byte
    for (lpos = 0; ((pos < maxsz) && (*(buf + pos))); ++pos, ++lpos)
        *(buf + lpos) = *(buf + pos);
    //lpos will always be < maxsz
    *(buf + lpos) = CHR_NULL;
    return(lpos);
} //End function strlnshift


/*Function strlrtrim.  Trims leading whitespace and all whitespace & text after
 *  first non-whitespace encounter.  Useful to isolate the leading text in a
 *  command string.
 *
 * Returns:     length of trimmed string or zero if not trimmed
 *
 * Parameters:  buf=pointer to string
 *              maxsz=size of buffer
 *
 */
int strlrtrim(unsigned char *buf, int maxsz)
{
    int pos,
        txt_s,                                  //Starting position
        txt_e,                                  //Ending position
        txt_f;                                  //Found start flag
    char chr;                                   //Byte assignment

    //
    for (pos = 0, txt_s = 0, txt_e = 0, txt_f = FALSE; \
         ((pos < maxsz) && (*(buf + pos))); ++pos)
    {
        if ((*(buf + pos) != CHR_SPACE) && (!(txt_f)))
        {
            txt_s = pos;
            txt_f = TRUE;
        }
        if ((*(buf + pos) == CHR_SPACE) && (txt_f)) break;
    }
    if (txt_f) txt_e = (pos - 1);

    //Reset pos so if nothing to trim tell caller; otherwise with length.
    //Necessarily inefficent loop to avoid same source/dest address when there
    //isn't leading whitespace.
    pos = 0;
    if (txt_f)
    {
        for (pos = 0; (txt_s <= txt_e); ++txt_s, ++pos)
        {
            chr = *(buf + txt_s);
            *(buf + pos) = chr;
        }
        *(buf + pos) = CHR_NULL;                //NULL terminate
    }

    return(pos);
} //End function strlrtrim


/*Function sigHandler.  Signal handler for user interrupt & TERM signals.
 * Attempts to ensure that open files / devices are closed.
 *
 * Returns:     n/a
 *
 * Parameters:  signal=signal #; refer to signal.h.
 *
 */
void sigHandler(int signal)
{
    static int recurse = 1;

    if (recurse)
    {
        recurse = 0;
        if (GpibDev[ARY_FLD_DEV1])
        {
            gpibOff(Index[DICT_OFF_F]);         //This call handles both devices
            fclose(GpibDev[ARY_FLD_DEV1]);
        }
        if (GpibDev[ARY_FLD_DEV2]) fclose(GpibDev[ARY_FLD_DEV2]);
        if (LogFile) fclose(LogFile);

        fprintf(stderr, FMT_TRM_SIGNAL, prgName, signal);
        exit(RC_SIGEXIT);
    }

    return;
} //End function sigHandler


/*Function getCfgVar.  Parses a config record for the given variable.  Not
 * thread safe! Modified in version 5.1 to allow escaping of semicolon within
 * a field to be part of instrument command syntax.
 *
 * Returns:     pointer to NULL terminated string; will be 0 length if variable
 *              not found, 0 length, or corrupt record
 *
 * Parameters:  rec=pointer to NULL terminated config record
 *              pos=offset-1 number representing variable position; if set to
 *              0 then the record identifier & first variable will be returned
 *
 */
const char *getCfgVar(char *rec, int pos)
{
    static char var[SIZ_BUF_VAR];
    int rec_off,
        var_off,
        pos_cur,
        fndc;

    memset((void *) var, CHR_NULL, SIZ_BUF_VAR);        //Clear
    for (rec_off = 0, var_off = 0, pos_cur = 0, fndc = 0; \
         (*(rec + rec_off)); ++rec_off)
    {
        if (pos_cur == pos)                             //Field to operate on
        {
            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
            * Append characters to field buffer while testing for:
            *   + End of record (NULL) - return
            *   + Full buffer - field in config record larger than allowed -
            *     return
            *   + End of field character (semicolon) - this can be escaped
            *     using the backslash character to allow use as part of
            *     instrument command string. If not escaped it is interpreted
            *     as the End Of Field (EOF) character.
            ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
            if (*(rec + rec_off) != CHR_NULL)           //EOF if NULL
            {
                if (*(rec + rec_off) == CHR_SEMICOLON)  //Test for EOF
                {
                    //If escaped then replace with semicolon
                    if (*(rec + rec_off - 1) == '\\')
                        var[--var_off] = *(rec + rec_off);
                    else break;                         //Otherwise EOF
                }
                else var[var_off] = *(rec + rec_off);   //Append to buffer
                if (++var_off < (SIZ_BUF_VAR - 1)) continue;
            }
            break;
        }

        //No field scanning until terminator for record identifier found
        if ((*(rec + rec_off) == CHR_COLON) && (!(fndc)))
        {
            fndc = TRUE;
            ++pos_cur;                                  //Found 1st variable
            continue;
        }

        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
        * Once record identifier found scan for requested field but skip skip
        * escaped semicolons: these are field string text and not delimiters.
        ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        if ((*(rec + rec_off) == CHR_SEMICOLON) && (fndc))
        {
            if (*(rec + rec_off - 1) != '\\')
                ++pos_cur;                              //Found next variable
        }
    }

    return((const char *) var);
} //End function getCfgVar


/*Function gpibDsp.  Handles all acquisition & rendering of display record
 * data.  For conventional fields reads values from GPIB device and displays
 * using DISP dictionary records.  Also processes calculated fields.
 *
 * At the start of each display update a progress character to the left of the
 * command prompt is updated so the user knows the display is not frozen.  If
 * the display reset flags (flagDspForce, flagDspReset) are set then the streams
 *  will also be purged of any unprocessed data at the start of each display
 * loop.
 *
 * Data logging is performed as part of display process if enabled.
 *
 * Runs as a thread.
 *
 * Returns:     n/a
 *
 * Parameters:  args=pointer to argument passed by parent
 *
 */
void gpibDsp(void *args)
{
    char *udchr = STR_UPDATE,
         *fgets_p;
    gcmd *dsp;
    int dsprow,
        udchr_l,
        udchr_p,
        loop_c,
        read_fail;
    unsigned int rec_num = 0;
    struct timespec base_to;
    time_t now;
    struct tm *now_tm;

    //Set base device timeout
    base_to.tv_sec = DLY_S_DUP_C;
    base_to.tv_nsec =  DLY_NS_RD_C;

    //These control the display of progress character
    udchr_p = 0;
    udchr_l = strlen(udchr);

    //No display reset regardless of flags at loop entry.
    read_fail = FALSE;

    while (TRUE)
    {
        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * Simple process control.  If gpibDspWait is non-zero set gpibDspAck
         * non-zero and halt all I/O.  When gpibDspWait clears, clear gpibDspAck
         * and resume.
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        if ((gpibDspWait) && (!(gpibDspAck))) gpibDspAck = TRUE;
        else if (!(gpibDspWait) && ((gpibDspAck))) gpibDspAck = FALSE;
        else if (!(gpibDspWait) && (!(gpibDspAck)))
        {
            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             * Flag read_fail controls the purging of any extraneous data in
             * the stream at the start of the display loop.  Command flags
             * flagDspForce & flagDspReset determine if this happens on every
             * display loop or only those where gpibRead fails, respectively.
             *
             * Data isn't read from the device stream; it is purged.  Each
             * stream with a valid handle is purged.  This does not guarantee
             * that extranesous data may still show-up in the stream if there
             * is extreme latency from the instrument.  This approach also has
             * the side effect of purging any unsent data due to blocking from
             * the instrument.
             +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
             if (read_fail)
             {
                 for (loop_c = 0; (loop_c < SIZ_ARY_DEV); ++loop_c)
                 {
                     if (GpibDev[loop_c]) fpurge(GpibDev[loop_c]);
                 }
             }

             /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
              * If logging get date & time and write new row
              *
              * If gpibDspMarkText contains a string write it since it is the
              * beginning of a mark on / off set.  Always set to NULL afterwards
              * so the data isn't repeated in subsequent records.
              +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
            if (LogFile)
            {
                now = time(NULL);
                now_tm = localtime(&now);
                fprintf(LogFile, FMT_LOG_OUT, ++rec_num, (now_tm->tm_mon + 1), \
                        now_tm->tm_mday, (now_tm->tm_year + SYS_EPOCH_YEAR), \
                        now_tm->tm_hour, now_tm->tm_min, now_tm->tm_sec, \
                        gpibDspMark, gpibDspMarkText);
                gpibDspMarkText[0] = CHR_NULL;
            }

            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             * Read GPIB data
             +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
            for (read_fail = flagDspForce, dsp = Index[DICT_DISP_F]; (dsp); \
                dsp = dsp->next)
            {
                if (dsp->math)
                {
                    //Calculated field
                    gpibCalc(dsp);
                }
                else
                {
                    //Set device address
                    setDevAddr(dsp);

                    //GPIB field, send command
                    if (dsp->devf) fprintf(dsp->devf, FMT_DSP_WR, dsp->cmd);
                    nanosleep(&base_to, NULL);

                    //Only issue read command if defined
                    if (strlen(Read_gpib[dsp->devn].cmd))
                        if (dsp->devf) fprintf(dsp->devf, FMT_DSP_WR, \
                                               Read_gpib[dsp->devn].cmd);

                    //Read data, follow leading text strip per crsp in object
                    if (!(gpibRead(dsp, dsp->crsp))) read_fail = flagDspReset;
                }
            }

            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             * Display GPIB data.  Flush stdout & save current cursor position.
             * Then update mark status before field display.
             +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
            fflush(stdout);
            flockfile(stdout);
            ansiCursorSav();

            //Update mark status
            ansiCursorPos(DSP_POS_MKS_R, DSP_POS_MKS_C);
            fprintf(stdout, FMT_MKS_MARK, ((gpibDspMark) ? STR_ON : STR_OFF));

            //Begin display field update
            ansiCursorPos(DSP_POS_UPD_R, DSP_POS_UPD_C);
            fprintf(stdout, FMT_STR_UPDATE, *(udchr + udchr_p));
            if (++udchr_p == udchr_l) udchr_p = 0;
            ansiCursorPos(DSP_POS_VAL_R, DSP_POS_VAL_C);
            fflush(stdout);
            for (dsprow = 1, dsp = Index[DICT_DISP_F]; (dsp); dsp = dsp->next)
            {
                //V4.2:  only display if row != 0
                if (dsp->usrv)
                {
                    if (dsprow != dsp->usrv)
                    {
                        dsprow = dsp->usrv;
                        ansiCursorPos((unsigned char) dsprow, DSP_POS_VAL_C);
                        ansiEraseLine();
                    }
                    fprintf(stdout, FMT_DSP_VAL, dsp->cfg, dsp->rsp);
                }

                //Write field if logging active
                if (LogFile) fprintf(LogFile, FMT_LOG_FLD, dsp->rsp);
            }
            ansiCursorRst();
            fflush(stdout);
            funlockfile(stdout);
        }

        //Sleep until next display update
        nanosleep(&update_tm, NULL);
    }

    return;
} //End function gpibDsp


/*Function ansiClrScr.  Clears terminal screen.
 *
 * Returns:     n/a
 *
 * Parameters:  none
 *
 */
void ansiClrScr(void)
{
    printf("%s%s", ANSIESC, ANSICLRSCR);
    return;
} //End function ansiClrScr


/*Function ansiEraseLine.  Erases line fromo cursor position to end of line.
 *
 * Returns:     n/a
 *
 * Parameters:  none
 *
 */
void ansiEraseLine(void)
{
    printf("%s%c", ANSIESC, ANSIELINE);
    return;
} //End function ansiEraseLine


/*Function ansiCursorSav.  Tells terminal to save terminal position.
 *
 * Returns:     n/a
 *
 * Parameters:  none
 *
 */
void ansiCursorSav(void)
{
    printf(ANSISAVE);
    return;
} //End function ansiCursorSav


/*Function ansiCursorRst.  Tells terminal to restore terminal position.
 *
 * Returns:     n/a
 *
 * Parameters:  none
 *
 */
void ansiCursorRst(void)
{
    printf(ANSIRESTORE);
    return;
} //End function ansiCursorRst


/*Function ansiCursorPos.  Sets terminal cursor position.
 *
 * Returns:     n/a
 *
 * Parameters:  row=terminal row
 *              col=terminal column
 *
 */
void ansiCursorPos(unsigned char row, unsigned char col)
{
    char srow[ACP_SZ_STR], scol[ACP_SZ_STR];

    if ((row < ACP_MAX_ROW) && (col < ACP_MAX_COL))
    {
        snprintf(srow, ACP_SZ_STR, "%hhu", row);
        snprintf(scol, ACP_SZ_STR, "%hhu", col);
        printf("%s%s;%s%c", ANSIESC, srow, scol, ANSICPOS);
    }
    return;
} //End function ansiCursorPos


/*Function scrSetup.  Sets up the console display.  If logging is active a
 * message with the file name is printed on a dedicated row below the command
 * entry row.
 *
 * If called with the display thread active the caller is expected to pause
 * display updates.
 *
 * Returns:     n/a
 *
 * Parameters:  n/a
 *
 */
void scrSetup(void)
{
    fflush(stdout);
    ansiClrScr();
    ansiCursorPos(DSP_POS_MKL_R, DSP_POS_MKL_C);
    fprintf(stdout, FMT_MKL_MARK);
    ansiCursorPos(DSP_POS_LOG_R, DSP_POS_LOG_C);
    if (LogFile) fprintf(stdout, FMT_LOG_MSG, Flags[FLG_A_LOG_FILE].value);
    ansiCursorPos(DSP_POS_PRG_R, DSP_POS_PRG_C);
    fprintf(stdout, FMT_PRG_DESC, VERMAJOR, VERMINOR, VERBUILD);
    fflush(stdout);

    return;
} //End function scrSetup


/*Function gpibCalc.  Calculates the value for a calculated field.  Fetches the
 * fld1 & fld2 values from those indicated in the record pointed to by calc,
 * converts them to type float, and applies the operand indicated by the math
 * field.  The result is converted to a string & stored in the rsp field.
 *
 * Constants are identified by calc->fld array element set to zero.  The
 * function then retrieves the value to be used from the calc->fval array
 * element.
 *
 * The result of the math operation is stored in calc->rsp as a string.
 *
 * Returns:     n/a
 *
 * Parameters:  calc=pointer to gcmd object to store result in (element rsp)
 *
 */
void gpibCalc(gcmd *calc)
{
    gcmd  *dsp;
    int   fcnt,                                 //Field counter
          fary;                                 //Object fld array pointer
    float fld[SIZ_ARY_FLD],                     //Array of float value of fields
          result;                               //Float value of result

    fld[ARY_FLD_FLD1] = 0;
    fld[ARY_FLD_FLD2] = 0;

    //Fetch GPIB source data & convert to float
    for (fary = SIZ_ARY_FLD, fcnt = 1, dsp = Index[DICT_DISP_F]; (dsp); ++fcnt, \
         dsp = dsp->next, fary = SIZ_ARY_FLD)
    {
        if (fcnt == calc->fld[ARY_FLD_FLD1]) fary = ARY_FLD_FLD1;
        else if (fcnt == calc->fld[ARY_FLD_FLD2]) fary = ARY_FLD_FLD2;
        if (fary < SIZ_ARY_FLD) fld[fary] = strtof(dsp->rsp, NULL);
    }

    //Assign constants if found:  indicated by field #=0
    if (!(calc->fld[ARY_FLD_FLD1])) fld[ARY_FLD_FLD1] = calc->fval[ARY_FLD_FLD1];
    if (!(calc->fld[ARY_FLD_FLD2])) fld[ARY_FLD_FLD2] = calc->fval[ARY_FLD_FLD2];

    //Perform calculation
    if (calc->math == MATH_OP_ADD) result = fld[ARY_FLD_FLD1] + \
        fld[ARY_FLD_FLD2];
    else if (calc->math == MATH_OP_SUB) result = fld[ARY_FLD_FLD1] - \
        fld[ARY_FLD_FLD2];
    else if (calc->math == MATH_OP_MUL) result = fld[ARY_FLD_FLD1] * \
        fld[ARY_FLD_FLD2];
    else if (calc->math == MATH_OP_DIV) result = fld[ARY_FLD_FLD1] / \
        fld[ARY_FLD_FLD2];

    //Store result
    snprintf(calc->rsp, SIZ_BUF_RSP, FMT_MTH_RESULT, calc->prec, result);

    return;
} //End function gpibCalc


/*Function setDevAddr.  Sets the active GPIB address for a device.  This
 * function works by using the respective address assigned to the instrument
 * (dictionary object) to set the current GPIB logical address within the
 * device.  If the address is already considered active then no command is
 * issued:  this saves time which can be important when issuing a large number
 * of commands in succession.
 *
 * This function expects that the address change command (provided by the user
 * in the config file) ends with the address number.
 *
 * Returns:     non-zero if successful including if the address for the
 *              specified device is already active.
 *
 * Parameters:  addr_r points to dictionary object containing valid device &
 *              address.
 *
 */
int setDevAddr(gcmd *addr_r)
{
    int rc = TRUE;
    static unsigned int addr_c[SIZ_ARY_ADR] = {VAL_ADR_INIT, VAL_ADR_INIT};

    if (addr_c[addr_r->devn] != addr_r->addr)
    {
        //Only attempt to change address if command defined & device open
        if ((strlen(Addr_gpib[addr_r->devn].cmd)) && (addr_r->devf))
            rc = fprintf(addr_r->devf, FMT_ADR_CHG, \
                         Addr_gpib[addr_r->devn].cmd, addr_r->addr);
        if (rc) addr_c[addr_r->devn] = addr_r->addr;
    }

    return(rc);
} //End function setDevAddr


/*Function gpibRead.  Reads stream data from GPIB device until newline or
 * timeout is encountered.
 *
 * Returns:     non-zero if valid line read from device (meaning 1 or more
 *              bytes ending with newline (LF) character.  If zero an I/O error
 *              occured or timeout value exceeded before complete line received.
 *              The contents of the rsp field in object gpib_d will be
 *              indeterminate.
 *
 * Parameters:  gpib_d points to dictionary object containing valid device
 *              stream to read from.
 *              strip_lt non-zero to remove leading text from response.  This
 *              parameter is provided versus using the crsp field for
 *              versatility since different functions in the application use
 *              this field differently.
 *
 * Known bugs:  responses exceeding the size of SIZ_BUF_RSP will return 0
 *              on the initial call and return the remainder of the data
 *              (providing the underlying OS, device, instrument, etc., does
 *              not discard) in a subsequent call with a success code.
 *
 *              The same behavior may occur if data is slow to arrive - e.g.
 *              it exceeds the timeout (usrv) value:  data may be broken into
 *              two or more successive reads.
 *
 */
int gpibRead(gcmd *gpib_d, int strip_lt)
{
    static int loop_c;
    static char rsp[SIZ_BUF_RSP],
                *dsperr = STR_ERROR,
                *fgets_p;
    static struct timespec base_to;

    //Set base device timeout
    base_to.tv_sec = DLY_S_DUP_C;
    base_to.tv_nsec =  DLY_NS_RD_C;

    //Clear return response
    *gpib_d->rsp = CHR_NULL;

    //Read loop until newline or timeout only if device is open
    if (gpib_d->devf)
    {
        for (loop_c = Read_gpib[gpib_d->devn].usrv; (loop_c); \
             --loop_c, nanosleep(&base_to, NULL))
        {
            if (fgets(rsp, SIZ_BUF_RSP, gpib_d->devf))
            {
                /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
                * v4.1: on Linux (at least Ubuntu) systems the stream appears to
                * add an extra LF.  This is removed to prevent processing as a
                * 'NULL' result (the data comes in the next read).
                *
                * To ensure that a legitimate LF isn't discarded (e.g. a
                * fragmented read returning the LF as the last read in
                * sequence), the response is appended and then evaluated for
                * only containing an LF.  If so, rsp is NULLed and the read
                * loop continues.
                *+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
                strlcat(gpib_d->rsp, rsp, SIZ_BUF_RSP);
                if (*(gpib_d->rsp + (strlen(gpib_d->rsp) - 1)) == CHR_LF)
                {
                    if (strlen(gpib_d->rsp) == 1) *gpib_d->rsp = CHR_NULL;
                    else break;
                }
            }
        }
    }
    else loop_c = 0;

    //Read valid line
    if (loop_c)
    {
        if (strip_lt) strlnshift((unsigned char *) gpib_d->rsp, SIZ_BUF_RSP);
        strnl((unsigned char *) gpib_d->rsp, SIZ_BUF_RSP);  //Remove LF
    }
    //Timeout or I/O error - set error string
    else strlcpy(gpib_d->rsp, dsperr, SIZ_BUF_RSP);

    return(loop_c);
} //End function gpibRead


/*Function gpibLogc.  Logs config (console) commands.  It is implemented with
 * no module dependencies.  On each call the file pointed to by fp is opened
 * for append writing, the record is writen, and then closed.  While
 * somewhat less efficient than keeping open throughout execution, this is
 * considered a reasonable trade-off since it is not expected that a high
 * volume / speed of writes will be done.
 *
 * If called with cmd=NULL, the file pointed to by fp will be initialized with
 * a header record.  If the file does not exist it will be created & the header
 * writen; otherwise the header is only writen if the file size is less than
 * the string length of the header.  This is done on the assumption that the
 * file was used in a previous session and already contains a header.
 *
 * When cmd is not NULL then a record is appended to the file that contains the
 * record number, date, time, command issued (cmd), command parameter (value),
 * and instrument response (rsp).
 *
 * Returns:     TRUE if file created or record written,
 *              FALSE on error.
 *
 * Parameters:  fp=pointer to NULL terminated file path
 *              cmd=pointer to NULL terminated command string; if NULL the
 *                  file fp will be initialized
 *              value=pointer to NULL terminated value string
 *              rsp=pointer to NULL terminated instrument response
 *
 */
int gpibLogc(char *fp, char *cmd, char *value, char *rsp)
{
    static FILE *stream_fp;
    static struct stat stat_fp;
    static time_t now;
    static struct tm *now_tm;
    static int rc = FALSE;
    static char *header_fp = FMT_CFG_HDR;
    static unsigned int rec_num = 0;

    //Waterfall processing loop
    for (rc = FALSE; (!(rc)); )
    {
        //Open file for append, create if doesn't exist
        if (!(stream_fp = fopen(fp, FMODE_CFG_FIL))) break;

        //Only write header if cmd=NULL and file size < header size
        if (!(cmd))
        {
            if (fstat(fileno(stream_fp), &stat_fp)) break;
            if (stat_fp.st_size < strlen(header_fp))
                if (!(fprintf(stream_fp, header_fp))) break;

            //Set return code TRUE provided no error
            rc = TRUE;
        }
        else
        {
            //Get current system time & write record.
            now = time(NULL);
            now_tm = localtime(&now);
            if (fprintf(stream_fp, FMT_CFG_OUT, ++rec_num, (now_tm->tm_mon + 1), \
                now_tm->tm_mday, (now_tm->tm_year + SYS_EPOCH_YEAR), \
                now_tm->tm_hour, now_tm->tm_min, now_tm->tm_sec, cmd, value,
                rsp) > 0) rc = TRUE;
        }

       //Always exit loop:  rc contains return code.
       break;
    }

    if (stream_fp) fclose(stream_fp);

   return(rc);
} //End function gpibLogc



//End of file.
